header {* Defining Red-Black Trees *}


theory 
  RBT_def
imports
  Testing
begin




(******************************)

  (* mimicking ml's datatype order *)
text {*
 The implementation of Red-Black
trees is mainly based on the following datatype declaration:
*}
datatype ml_order = LESS | EQUAL | GREATER  

axclass ord_key < type      

consts
  compare :: "'a::ord_key \<Rightarrow> 'a \<Rightarrow> ml_order "  

axclass LINORDER < linorder, ord_key
  LINORDER_less    : "((compare x y) = LESS)    = (x < y)"
  LINORDER_equal   : "((compare x y) = EQUAL)   = (x = y)"
  LINORDER_greater : "((compare x y) = GREATER) = (y < x)"

(* defining red-black trees *)

types    'a item = "'a::ord_key" 

datatype  color = R | B

datatype 'a tree = E | T color "'a tree" "'a item" "'a tree"


  
(* invariants *)

text {*
In this example we have chosen not only to check if keys are stored or
deleted correctly in the trees but also to check if the trees fulfill
the balancing invariants. We formalize the red and black invariant by recursive
predicates:


*}

consts
  isin         :: "'a::LINORDER item \<Rightarrow> 'a tree \<Rightarrow> bool"
  isord        :: "('a::LINORDER item) tree \<Rightarrow> bool"
  redinv       :: "'a tree \<Rightarrow> bool"
  blackinv     :: "'a tree \<Rightarrow> bool"
  strong_redinv:: "'a tree \<Rightarrow> bool"
  max_B_height :: "'a tree \<Rightarrow> nat"

(* test for elements *)
primrec
  isin_empty : "isin x E = False"
  isin_branch: "isin x (T c a y b) = (((compare x y) = EQUAL) 
                                     | (isin x a) | (isin x b))"

(* test for order of elements *) 
primrec
  isord_empty : "isord E = True"
  isord_branch: "isord (T c a y b) 
                = (isord a \<and> isord b 
                \<and> (\<forall> x. isin x a \<longrightarrow> ((compare x y) = LESS))
                \<and> (\<forall> x. isin x b \<longrightarrow> ((compare x y) = GREATER)))"

(* weak red invariant: no red node has a red child *)  
recdef redinv "measure (%t. (size t))"
  redinv_1:   "redinv E = True"
  redinv_2:   "redinv (T B a y b) = (redinv a \<and> redinv b)"
  redinv_3:   "redinv (T R (T R a x b) y c) = False"
  redinv_4:   "redinv (T R a x (T R b y c)) = False"
  redinv_5:   "redinv (T R a x b) = (redinv a \<and> redinv b)"

(* strong red invariant: every red node has an immediate black ancestor, i.e. 
   the root is black and the weak red invariant holds *)
recdef strong_redinv "{}"
  Rinv_1:  "strong_redinv E = True"
  Rinv_2:  "strong_redinv (T R a y b) = False"
  Rinv_3:  "strong_redinv (T B a y b) = (redinv a \<and> redinv b)"

(* calculating maximal number of black nodes on any path from root to leaf *)
recdef max_B_height "measure (%t. (size t))"
  maxB_height_1:  "max_B_height E = 0"
  maxB_height_3:  "max_B_height (T B a y b) 
                   = Suc(max (max_B_height a) (max_B_height b))"
  maxB_height_2:  "max_B_height (T R a y b) 
                   = (max (max_B_height a) (max_B_height b))"

(* black invariant: number of black nodes equal on all paths from root to leaf *)  
recdef blackinv "measure (%t. (size t))"
  blackinv_1:  "blackinv E = True"
  blackinv_2:  "blackinv (T color a y b) 
               = ((blackinv a) \<and> (blackinv b) 
               \<and> ((max_B_height a) = (max_B_height b)))"
instance int::ord_key
  by(intro_classes)

instance int::linorder
  by intro_classes

defs compare_def: "compare (x::int) y 
                  == (if (x < y) then LESS 
                                 else (if (y < x) 
                                       then GREATER 
                                       else EQUAL))"

instance int::LINORDER
apply intro_classes
apply (simp_all add: compare_def)
done


lemma compare1[simp]:"(compare (x::int) y = EQUAL) = (x=y)"
by(auto simp:compare_def)

lemma compare2[simp]:"(compare (x::int) y = LESS) = (x<y)"
by(auto simp:compare_def)

lemma compare3[simp]:"(compare (x::int) y = GREATER) = (y<x)"
by(auto simp:compare_def)


lemma max_0_0 : "((max (a::nat) b) = 0) = (a = 0 \<and> (b = 0))"
  by(auto simp: max_def)

lemma [simp]: "(max (Suc a) b) ~= 0"
  by(auto simp: max_def)

lemma [simp]: "(max b (Suc a) ) ~= 0"
  by(auto simp: max_def)

lemma size_0[simp]: "(size x = 0) = (x = E)"
  by(induct "x",auto)




end
